library angular2.src.compiler.template_ast;

import "package:angular2/src/core/change_detection/change_detection.dart"
    show AST;
import "package:angular2/src/core/facade/lang.dart" show isPresent;
import "api.dart" show DirectiveMetadata;

abstract class TemplateAst {
  String sourceInfo;
  dynamic visit(TemplateAstVisitor visitor);
}

class TextAst implements TemplateAst {
  String value;
  String sourceInfo;
  TextAst(this.value, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitText(this);
  }
}

class BoundTextAst implements TemplateAst {
  AST value;
  String sourceInfo;
  BoundTextAst(this.value, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitBoundText(this);
  }
}

class AttrAst implements TemplateAst {
  String name;
  String value;
  String sourceInfo;
  AttrAst(this.name, this.value, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitAttr(this);
  }
}

class BoundElementPropertyAst implements TemplateAst {
  String name;
  PropertyBindingType type;
  AST value;
  String unit;
  String sourceInfo;
  BoundElementPropertyAst(
      this.name, this.type, this.value, this.unit, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitElementProperty(this);
  }
}

class BoundEventAst implements TemplateAst {
  String name;
  String target;
  AST handler;
  String sourceInfo;
  BoundEventAst(this.name, this.target, this.handler, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitEvent(this);
  }
}

class VariableAst implements TemplateAst {
  String name;
  String value;
  String sourceInfo;
  VariableAst(this.name, this.value, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitVariable(this);
  }
}

class ElementAst implements TemplateAst {
  List<AttrAst> attrs;
  List<BoundElementPropertyAst> properties;
  List<BoundEventAst> events;
  List<VariableAst> vars;
  List<DirectiveAst> directives;
  List<TemplateAst> children;
  String sourceInfo;
  ElementAst(this.attrs, this.properties, this.events, this.vars,
      this.directives, this.children, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitElement(this);
  }
}

class EmbeddedTemplateAst implements TemplateAst {
  List<AttrAst> attrs;
  List<VariableAst> vars;
  List<DirectiveAst> directives;
  List<TemplateAst> children;
  String sourceInfo;
  EmbeddedTemplateAst(
      this.attrs, this.vars, this.directives, this.children, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitEmbeddedTemplate(this);
  }
}

class BoundDirectivePropertyAst implements TemplateAst {
  String directiveName;
  String templateName;
  AST value;
  String sourceInfo;
  BoundDirectivePropertyAst(
      this.directiveName, this.templateName, this.value, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitDirectiveProperty(this);
  }
}

class DirectiveAst implements TemplateAst {
  DirectiveMetadata directive;
  List<BoundDirectivePropertyAst> properties;
  List<BoundElementPropertyAst> hostProperties;
  List<BoundEventAst> hostEvents;
  String sourceInfo;
  DirectiveAst(this.directive, this.properties, this.hostProperties,
      this.hostEvents, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitDirective(this);
  }
}

class NgContentAst implements TemplateAst {
  String select;
  String sourceInfo;
  NgContentAst(this.select, this.sourceInfo) {}
  dynamic visit(TemplateAstVisitor visitor) {
    return visitor.visitNgContent(this);
  }
}

enum PropertyBindingType { Property, Attribute, Class, Style }

abstract class TemplateAstVisitor {
  dynamic visitNgContent(NgContentAst ast);
  dynamic visitEmbeddedTemplate(EmbeddedTemplateAst ast);
  dynamic visitElement(ElementAst ast);
  dynamic visitVariable(VariableAst ast);
  dynamic visitEvent(BoundEventAst ast);
  dynamic visitElementProperty(BoundElementPropertyAst ast);
  dynamic visitAttr(AttrAst ast);
  dynamic visitBoundText(BoundTextAst ast);
  dynamic visitText(TextAst ast);
  dynamic visitDirective(DirectiveAst ast);
  dynamic visitDirectiveProperty(BoundDirectivePropertyAst ast);
}

List<dynamic> templateVisitAll(
    TemplateAstVisitor visitor, List<TemplateAst> asts) {
  var result = [];
  asts.forEach((ast) {
    var astResult = ast.visit(visitor);
    if (isPresent(astResult)) {
      result.add(astResult);
    }
  });
  return result;
}
