library angular2.src.compiler.template_loader;

import "api.dart" show TypeMetadata, TemplateMetadata, ViewEncapsulation;
import "package:angular2/src/core/facade/lang.dart" show isPresent;
import "package:angular2/src/core/facade/async.dart"
    show Future, PromiseWrapper;
import "package:angular2/src/core/render/xhr.dart" show XHR;
import "package:angular2/src/core/services/url_resolver.dart" show UrlResolver;
import "style_url_resolver.dart" show StyleUrlResolver;
import "html_ast.dart"
    show
        HtmlAstVisitor,
        HtmlElementAst,
        HtmlTextAst,
        HtmlAttrAst,
        HtmlAst,
        htmlVisitAll;
import "html_parser.dart" show HtmlParser;

const NG_CONTENT_SELECT_ATTR = "select";
const NG_CONTENT_ELEMENT = "ng-content";
const LINK_ELEMENT = "link";
const LINK_STYLE_REL_ATTR = "rel";
const LINK_STYLE_HREF_ATTR = "href";
const LINK_STYLE_REL_VALUE = "stylesheet";
const STYLE_ELEMENT = "style";

class TemplateLoader {
  XHR _xhr;
  UrlResolver _urlResolver;
  StyleUrlResolver _styleUrlResolver;
  HtmlParser _domParser;
  TemplateLoader(
      this._xhr, this._urlResolver, this._styleUrlResolver, this._domParser) {}
  Future<TemplateMetadata> loadTemplate(
      TypeMetadata directiveType,
      ViewEncapsulation encapsulation,
      String template,
      String templateUrl,
      List<String> styles,
      List<String> styleUrls) {
    if (isPresent(template)) {
      return PromiseWrapper.resolve(this.createTemplateFromString(directiveType,
          encapsulation, template, directiveType.typeUrl, styles, styleUrls));
    } else {
      var sourceAbsUrl =
          this._urlResolver.resolve(directiveType.typeUrl, templateUrl);
      return this._xhr.get(sourceAbsUrl).then((templateContent) => this
          .createTemplateFromString(directiveType, encapsulation,
              templateContent, sourceAbsUrl, styles, styleUrls));
    }
  }

  TemplateMetadata createTemplateFromString(
      TypeMetadata directiveType,
      ViewEncapsulation encapsulation,
      String template,
      String templateSourceUrl,
      List<String> styles,
      List<String> styleUrls) {
    var domNodes = this._domParser.parse(template, directiveType.typeName);
    var visitor = new TemplatePreparseVisitor();
    var remainingNodes = htmlVisitAll(visitor, domNodes);
    var allStyles = new List.from(styles)..addAll(visitor.styles);
    var allStyleUrls = new List.from(styleUrls)..addAll(visitor.styleUrls);
    allStyles = allStyles.map((style) {
      var styleWithImports = this._styleUrlResolver.extractImports(style);
      styleWithImports.styleUrls
          .forEach((styleUrl) => allStyleUrls.add(styleUrl));
      return styleWithImports.style;
    }).toList();
    var allResolvedStyles = allStyles
        .map((style) =>
            this._styleUrlResolver.resolveUrls(style, templateSourceUrl))
        .toList();
    var allStyleAbsUrls = allStyleUrls
        .map((styleUrl) =>
            this._urlResolver.resolve(templateSourceUrl, styleUrl))
        .toList();
    return new TemplateMetadata(
        encapsulation: encapsulation,
        nodes: remainingNodes,
        styles: allResolvedStyles,
        styleAbsUrls: allStyleAbsUrls,
        ngContentSelectors: visitor.ngContentSelectors);
  }
}

class TemplatePreparseVisitor implements HtmlAstVisitor {
  List<String> ngContentSelectors = [];
  List<String> styles = [];
  List<String> styleUrls = [];
  HtmlElementAst visitElement(HtmlElementAst ast) {
    var selectAttr = null;
    var hrefAttr = null;
    var relAttr = null;
    ast.attrs.forEach((attr) {
      if (attr.name == NG_CONTENT_SELECT_ATTR) {
        selectAttr = attr.value;
      } else if (attr.name == LINK_STYLE_HREF_ATTR) {
        hrefAttr = attr.value;
      } else if (attr.name == LINK_STYLE_REL_ATTR) {
        relAttr = attr.value;
      }
    });
    var nodeName = ast.name;
    var keepElement = true;
    if (nodeName == NG_CONTENT_ELEMENT) {
      this.ngContentSelectors.add(selectAttr);
    } else if (nodeName == STYLE_ELEMENT) {
      keepElement = false;
      var textContent = "";
      ast.children.forEach((child) {
        if (child is HtmlTextAst) {
          textContent += ((child as HtmlTextAst)).value;
        }
      });
      this.styles.add(textContent);
    } else if (nodeName == LINK_ELEMENT && relAttr == LINK_STYLE_REL_VALUE) {
      keepElement = false;
      this.styleUrls.add(hrefAttr);
    }
    if (keepElement) {
      return new HtmlElementAst(ast.name, ast.attrs,
          htmlVisitAll(this, ast.children), ast.sourceInfo);
    } else {
      return null;
    }
  }

  HtmlAttrAst visitAttr(HtmlAttrAst ast) {
    return ast;
  }

  HtmlTextAst visitText(HtmlTextAst ast) {
    return ast;
  }
}
