library angular2.src.compiler.template_parser;

import "package:angular2/src/core/facade/collection.dart"
    show MapWrapper, ListWrapper, StringMapWrapper;
import "package:angular2/src/core/facade/lang.dart"
    show
        RegExpWrapper,
        isPresent,
        StringWrapper,
        BaseException,
        StringJoiner,
        stringify,
        assertionsEnabled,
        isBlank;
import "package:angular2/src/core/change_detection/change_detection.dart"
    show Parser, AST, ASTWithSource;
import "package:angular2/src/core/change_detection/parser/ast.dart"
    show TemplateBinding;
import "api.dart" show DirectiveMetadata;
import "template_ast.dart"
    show
        ElementAst,
        BoundElementPropertyAst,
        BoundEventAst,
        VariableAst,
        TemplateAst,
        TextAst,
        BoundTextAst,
        EmbeddedTemplateAst,
        AttrAst,
        NgContentAst,
        PropertyBindingType,
        DirectiveAst,
        BoundDirectivePropertyAst;
import "package:angular2/src/core/render/dom/compiler/selector.dart"
    show CssSelector, SelectorMatcher;
import "package:angular2/src/core/render/dom/schema/element_schema_registry.dart"
    show ElementSchemaRegistry;
import "html_ast.dart"
    show
        HtmlAstVisitor,
        HtmlAst,
        HtmlElementAst,
        HtmlAttrAst,
        HtmlTextAst,
        htmlVisitAll;
import "util.dart" show dashCaseToCamelCase, camelCaseToDashCase;
// Group 1 = "bind-"

// Group 2 = "var-" or "#"

// Group 3 = "on-"

// Group 4 = "bindon-"

// Group 5 = the identifier after "bind-", "var-/#", or "on-"

// Group 6 = idenitifer inside [()]

// Group 7 = idenitifer inside []

// Group 8 = identifier inside ()
var BIND_NAME_REGEXP = new RegExp(
    r'^(?:(?:(?:(bind-)|(var-|#)|(on-)|(bindon-))(.+))|\[\(([^\)]+)\)\]|\[([^\]]+)\]|\(([^\)]+)\))$');
const NG_CONTENT_SELECT_ATTR = "select";
const NG_CONTENT_ELEMENT = "ng-content";
const TEMPLATE_ELEMENT = "template";
const TEMPLATE_ATTR = "template";
const TEMPLATE_ATTR_PREFIX = "*";
const CLASS_ATTR = "class";
const IMPLICIT_VAR_NAME = "\$implicit";
var PROPERTY_PARTS_SEPARATOR = new RegExp("\\.");
const ATTRIBUTE_PREFIX = "attr";
const CLASS_PREFIX = "class";
const STYLE_PREFIX = "style";

class TemplateParser {
  Parser _exprParser;
  ElementSchemaRegistry _schemaRegistry;
  TemplateParser(this._exprParser, this._schemaRegistry) {}
  List<TemplateAst> parse(
      List<HtmlAst> domNodes, List<DirectiveMetadata> directives) {
    var parseVisitor = new TemplateParseVisitor(
        directives, this._exprParser, this._schemaRegistry);
    var result = htmlVisitAll(parseVisitor, domNodes);
    if (parseVisitor.errors.length > 0) {
      var errorString = parseVisitor.errors.join("\n");
      throw new BaseException('''Template parse errors:
${ errorString}''');
    }
    return result;
  }
}

class TemplateParseVisitor implements HtmlAstVisitor {
  Parser _exprParser;
  ElementSchemaRegistry _schemaRegistry;
  SelectorMatcher selectorMatcher;
  List<String> errors = [];
  TemplateParseVisitor(List<DirectiveMetadata> directives, this._exprParser,
      this._schemaRegistry) {
    this.selectorMatcher = new SelectorMatcher();
    directives.forEach((directive) {
      var selector = CssSelector.parse(directive.selector);
      this.selectorMatcher.addSelectables(selector, directive);
    });
  }
  _reportError(String message) {
    this.errors.add(message);
  }

  ASTWithSource _parseInterpolation(String value, String sourceInfo) {
    try {
      return this._exprParser.parseInterpolation(value, sourceInfo);
    } catch (e, e_stack) {
      this._reportError('''${ e}''');
      return this._exprParser.wrapLiteralPrimitive("ERROR", sourceInfo);
    }
  }

  ASTWithSource _parseAction(String value, String sourceInfo) {
    try {
      return this._exprParser.parseAction(value, sourceInfo);
    } catch (e, e_stack) {
      this._reportError('''${ e}''');
      return this._exprParser.wrapLiteralPrimitive("ERROR", sourceInfo);
    }
  }

  ASTWithSource _parseBinding(String value, String sourceInfo) {
    try {
      return this._exprParser.parseBinding(value, sourceInfo);
    } catch (e, e_stack) {
      this._reportError('''${ e}''');
      return this._exprParser.wrapLiteralPrimitive("ERROR", sourceInfo);
    }
  }

  List<TemplateBinding> _parseTemplateBindings(
      String value, String sourceInfo) {
    try {
      return this._exprParser.parseTemplateBindings(value, sourceInfo);
    } catch (e, e_stack) {
      this._reportError('''${ e}''');
      return [];
    }
  }

  dynamic visitText(HtmlTextAst ast) {
    var expr = this._parseInterpolation(ast.value, ast.sourceInfo);
    if (isPresent(expr)) {
      return new BoundTextAst(expr, ast.sourceInfo);
    } else {
      return new TextAst(ast.value, ast.sourceInfo);
    }
  }

  dynamic visitAttr(HtmlAttrAst ast) {
    return new AttrAst(ast.name, ast.value, ast.sourceInfo);
  }

  dynamic visitElement(HtmlElementAst element) {
    var nodeName = element.name;
    List<List<String>> matchableAttrs = [];
    List<BoundElementOrDirectiveProperty> elementOrDirectiveProps = [];
    List<VariableAst> vars = [];
    List<BoundEventAst> events = [];
    List<BoundElementOrDirectiveProperty> templateElementOrDirectiveProps = [];
    List<VariableAst> templateVars = [];
    List<List<String>> templateMatchableAttrs = [];
    var hasInlineTemplates = false;
    var attrs = [];
    var selectAttr = null;
    element.attrs.forEach((attr) {
      matchableAttrs.add([attr.name, attr.value]);
      if (attr.name == NG_CONTENT_SELECT_ATTR) {
        selectAttr = attr.value;
      }
      var hasBinding = this._parseAttr(
          attr, matchableAttrs, elementOrDirectiveProps, events, vars);
      var hasTemplateBinding = this._parseInlineTemplateBinding(
          attr,
          templateMatchableAttrs,
          templateElementOrDirectiveProps,
          templateVars);
      if (!hasBinding && !hasTemplateBinding) {
        // don't include the bindings as attributes as well in the AST
        attrs.add(this.visitAttr(attr));
      }
      if (hasTemplateBinding) {
        hasInlineTemplates = true;
      }
    });
    var directives = this._createDirectiveAsts(
        element.name,
        this._parseDirectives(this.selectorMatcher, nodeName, matchableAttrs),
        elementOrDirectiveProps,
        element.sourceInfo);
    List<BoundElementPropertyAst> elementProps = this
        ._createElementPropertyAsts(
            element.name, elementOrDirectiveProps, directives);
    var children = htmlVisitAll(this, element.children);
    var parsedElement;
    if (nodeName == NG_CONTENT_ELEMENT) {
      parsedElement = new NgContentAst(selectAttr, element.sourceInfo);
    } else if (nodeName == TEMPLATE_ELEMENT) {
      this._assertNoComponentsNorElementBindingsOnTemplate(
          directives, elementProps, events, element.sourceInfo);
      parsedElement = new EmbeddedTemplateAst(
          attrs, vars, directives, children, element.sourceInfo);
    } else {
      this._assertOnlyOneComponent(directives, element.sourceInfo);
      parsedElement = new ElementAst(attrs, elementProps, events, vars,
          directives, children, element.sourceInfo);
    }
    if (hasInlineTemplates) {
      var templateDirectives = this._createDirectiveAsts(
          element.name,
          this._parseDirectives(
              this.selectorMatcher, TEMPLATE_ELEMENT, templateMatchableAttrs),
          templateElementOrDirectiveProps,
          element.sourceInfo);
      List<BoundElementPropertyAst> templateElementProps = this
          ._createElementPropertyAsts(element.name,
              templateElementOrDirectiveProps, templateDirectives);
      this._assertNoComponentsNorElementBindingsOnTemplate(
          templateDirectives, templateElementProps, [], element.sourceInfo);
      parsedElement = new EmbeddedTemplateAst([], templateVars,
          templateDirectives, [parsedElement], element.sourceInfo);
    }
    return parsedElement;
  }

  bool _parseInlineTemplateBinding(
      HtmlAttrAst attr,
      List<List<String>> targetMatchableAttrs,
      List<BoundElementOrDirectiveProperty> targetProps,
      List<VariableAst> targetVars) {
    var templateBindingsSource = null;
    if (attr.name == TEMPLATE_ATTR) {
      templateBindingsSource = attr.value;
    } else if (StringWrapper.startsWith(attr.name, TEMPLATE_ATTR_PREFIX)) {
      var key = StringWrapper.substring(attr.name, TEMPLATE_ATTR_PREFIX.length);
      templateBindingsSource =
          (attr.value.length == 0) ? key : key + " " + attr.value;
    }
    if (isPresent(templateBindingsSource)) {
      var bindings =
          this._parseTemplateBindings(templateBindingsSource, attr.sourceInfo);
      for (var i = 0; i < bindings.length; i++) {
        var binding = bindings[i];
        var dashCaseKey = camelCaseToDashCase(binding.key);
        if (binding.keyIsVar) {
          targetVars.add(new VariableAst(
              dashCaseToCamelCase(binding.key), binding.name, attr.sourceInfo));
          targetMatchableAttrs.add([dashCaseKey, binding.name]);
        } else if (isPresent(binding.expression)) {
          this._parsePropertyAst(dashCaseKey, binding.expression,
              attr.sourceInfo, targetMatchableAttrs, targetProps);
        } else {
          targetMatchableAttrs.add([dashCaseKey, ""]);
        }
      }
      return true;
    }
    return false;
  }

  bool _parseAttr(
      HtmlAttrAst attr,
      List<List<String>> targetMatchableAttrs,
      List<BoundElementOrDirectiveProperty> targetProps,
      List<BoundEventAst> targetEvents,
      List<VariableAst> targetVars) {
    var attrName = this._normalizeAttributeName(attr.name);
    var attrValue = attr.value;
    var bindParts = RegExpWrapper.firstMatch(BIND_NAME_REGEXP, attrName);
    var hasBinding = false;
    if (isPresent(bindParts)) {
      hasBinding = true;
      if (isPresent(bindParts[1])) {
        this._parseProperty(bindParts[5], attrValue, attr.sourceInfo,
            targetMatchableAttrs, targetProps);
      } else if (isPresent(bindParts[2])) {
        var identifier = bindParts[5];
        var value =
            identical(attrValue.length, 0) ? IMPLICIT_VAR_NAME : attrValue;
        this._parseVariable(identifier, value, attr.sourceInfo,
            targetMatchableAttrs, targetVars);
      } else if (isPresent(bindParts[3])) {
        this._parseEvent(bindParts[5], attrValue, attr.sourceInfo,
            targetMatchableAttrs, targetEvents);
      } else if (isPresent(bindParts[4])) {
        this._parseProperty(bindParts[5], attrValue, attr.sourceInfo,
            targetMatchableAttrs, targetProps);
        this._parseAssignmentEvent(bindParts[5], attrValue, attr.sourceInfo,
            targetMatchableAttrs, targetEvents);
      } else if (isPresent(bindParts[6])) {
        this._parseProperty(bindParts[6], attrValue, attr.sourceInfo,
            targetMatchableAttrs, targetProps);
        this._parseAssignmentEvent(bindParts[6], attrValue, attr.sourceInfo,
            targetMatchableAttrs, targetEvents);
      } else if (isPresent(bindParts[7])) {
        this._parseProperty(bindParts[7], attrValue, attr.sourceInfo,
            targetMatchableAttrs, targetProps);
      } else if (isPresent(bindParts[8])) {
        this._parseEvent(bindParts[8], attrValue, attr.sourceInfo,
            targetMatchableAttrs, targetEvents);
      }
    } else {
      hasBinding = this._parsePropertyInterpolation(attrName, attrValue,
          attr.sourceInfo, targetMatchableAttrs, targetProps);
    }
    if (!hasBinding) {
      this._parseLiteralAttr(attrName, attrValue, attr.sourceInfo, targetProps);
    }
    return hasBinding;
  }

  String _normalizeAttributeName(String attrName) {
    return StringWrapper.startsWith(attrName, "data-")
        ? StringWrapper.substring(attrName, 5)
        : attrName;
  }

  _parseVariable(String identifier, String value, dynamic sourceInfo,
      List<List<String>> targetMatchableAttrs, List<VariableAst> targetVars) {
    targetVars.add(
        new VariableAst(dashCaseToCamelCase(identifier), value, sourceInfo));
    targetMatchableAttrs.add([identifier, value]);
  }

  _parseProperty(
      String name,
      String expression,
      dynamic sourceInfo,
      List<List<String>> targetMatchableAttrs,
      List<BoundElementOrDirectiveProperty> targetProps) {
    this._parsePropertyAst(name, this._parseBinding(expression, sourceInfo),
        sourceInfo, targetMatchableAttrs, targetProps);
  }

  bool _parsePropertyInterpolation(
      String name,
      String value,
      dynamic sourceInfo,
      List<List<String>> targetMatchableAttrs,
      List<BoundElementOrDirectiveProperty> targetProps) {
    var expr = this._parseInterpolation(value, sourceInfo);
    if (isPresent(expr)) {
      this._parsePropertyAst(
          name, expr, sourceInfo, targetMatchableAttrs, targetProps);
      return true;
    }
    return false;
  }

  _parsePropertyAst(
      String name,
      ASTWithSource ast,
      dynamic sourceInfo,
      List<List<String>> targetMatchableAttrs,
      List<BoundElementOrDirectiveProperty> targetProps) {
    targetMatchableAttrs.add([name, ast.source]);
    targetProps
        .add(new BoundElementOrDirectiveProperty(name, ast, false, sourceInfo));
  }

  _parseAssignmentEvent(
      String name,
      String expression,
      String sourceInfo,
      List<List<String>> targetMatchableAttrs,
      List<BoundEventAst> targetEvents) {
    this._parseEvent(name, '''${ expression}=\$event''', sourceInfo,
        targetMatchableAttrs, targetEvents);
  }

  _parseEvent(
      String name,
      String expression,
      String sourceInfo,
      List<List<String>> targetMatchableAttrs,
      List<BoundEventAst> targetEvents) {
    // long format: 'target: eventName'
    var parts = splitAtColon(name, [null, name]);
    var target = parts[0];
    var eventName = parts[1];
    targetEvents.add(new BoundEventAst(dashCaseToCamelCase(eventName), target,
        this._parseAction(expression, sourceInfo), sourceInfo));
  }

  _parseLiteralAttr(String name, String value, String sourceInfo,
      List<BoundElementOrDirectiveProperty> targetProps) {
    targetProps.add(new BoundElementOrDirectiveProperty(
        dashCaseToCamelCase(name),
        this._exprParser.wrapLiteralPrimitive(value, sourceInfo),
        true,
        sourceInfo));
  }

  List<DirectiveMetadata> _parseDirectives(SelectorMatcher selectorMatcher,
      String elementName, List<List<String>> matchableAttrs) {
    var cssSelector = new CssSelector();
    cssSelector.setElement(elementName);
    for (var i = 0; i < matchableAttrs.length; i++) {
      var attrName = matchableAttrs[i][0].toLowerCase();
      var attrValue = matchableAttrs[i][1];
      cssSelector.addAttribute(attrName, attrValue);
      if (attrName == CLASS_ATTR) {
        var classes = splitClasses(attrValue);
        classes.forEach((className) => cssSelector.addClassName(className));
      }
    }
    var directives = [];
    selectorMatcher.match(cssSelector, (selector, directive) {
      directives.add(directive);
    });
    // Need to sort the directives so that we get consistent results throughout,

    // as selectorMatcher uses Maps inside.

    // Also need to make components the first directive in the array
    ListWrapper.sort(directives,
        (DirectiveMetadata dir1, DirectiveMetadata dir2) {
      var dir1Comp = dir1.isComponent;
      var dir2Comp = dir2.isComponent;
      if (dir1Comp && !dir2Comp) {
        return -1;
      } else if (!dir1Comp && dir2Comp) {
        return 1;
      } else {
        return StringWrapper.compare(dir1.type.typeName, dir2.type.typeName);
      }
    });
    return directives;
  }

  List<DirectiveAst> _createDirectiveAsts(
      String elementName,
      List<DirectiveMetadata> directives,
      List<BoundElementOrDirectiveProperty> props,
      String sourceInfo) {
    return directives.map((DirectiveMetadata directive) {
      List<BoundElementPropertyAst> hostProperties = [];
      List<BoundEventAst> hostEvents = [];
      List<BoundDirectivePropertyAst> directiveProperties = [];
      var changeDetection = directive.changeDetection;
      if (isPresent(changeDetection)) {
        this._createDirectiveHostPropertyAsts(elementName,
            changeDetection.hostProperties, sourceInfo, hostProperties);
        this._createDirectiveHostEventAsts(
            changeDetection.hostListeners, sourceInfo, hostEvents);
        this._createDirectivePropertyAsts(
            changeDetection.properties, props, directiveProperties);
      }
      return new DirectiveAst(directive, directiveProperties, hostProperties,
          hostEvents, sourceInfo);
    }).toList();
  }

  _createDirectiveHostPropertyAsts(
      String elementName,
      Map<String, String> hostProps,
      String sourceInfo,
      List<BoundElementPropertyAst> targetPropertyAsts) {
    if (isPresent(hostProps)) {
      StringMapWrapper.forEach(hostProps, (expression, propName) {
        var exprAst = this._parseBinding(expression, sourceInfo);
        targetPropertyAsts.add(this._createElementPropertyAst(
            elementName, propName, exprAst, sourceInfo));
      });
    }
  }

  _createDirectiveHostEventAsts(Map<String, String> hostListeners,
      String sourceInfo, List<BoundEventAst> targetEventAsts) {
    if (isPresent(hostListeners)) {
      StringMapWrapper.forEach(hostListeners, (expression, propName) {
        this._parseEvent(propName, expression, sourceInfo, [], targetEventAsts);
      });
    }
  }

  _createDirectivePropertyAsts(
      List<String> directiveProperties,
      List<BoundElementOrDirectiveProperty> boundProps,
      List<BoundDirectivePropertyAst> targetBoundDirectiveProps) {
    if (isPresent(directiveProperties)) {
      Map<String, BoundElementOrDirectiveProperty> boundPropsByName = new Map();
      boundProps.forEach((boundProp) =>
          boundPropsByName[dashCaseToCamelCase(boundProp.name)] = boundProp);
      directiveProperties.forEach((String bindConfig) {
        // canonical syntax: `dirProp: elProp`

        // if there is no `:`, use dirProp = elProp
        var parts = splitAtColon(bindConfig, [bindConfig, bindConfig]);
        var dirProp = parts[0];
        var elProp = dashCaseToCamelCase(parts[1]);
        var boundProp = boundPropsByName[elProp];
        // Bindings are optional, so this binding only needs to be set up if an expression is given.
        if (isPresent(boundProp)) {
          targetBoundDirectiveProps.add(new BoundDirectivePropertyAst(dirProp,
              boundProp.name, boundProp.expression, boundProp.sourceInfo));
        }
      });
    }
  }

  List<BoundElementPropertyAst> _createElementPropertyAsts(
      String elementName,
      List<BoundElementOrDirectiveProperty> props,
      List<DirectiveAst> directives) {
    List<BoundElementPropertyAst> boundElementProps = [];
    Map<String, BoundDirectivePropertyAst> boundDirectivePropsIndex = new Map();
    directives.forEach((DirectiveAst directive) {
      directive.properties.forEach((BoundDirectivePropertyAst prop) {
        boundDirectivePropsIndex[prop.templateName] = prop;
      });
    });
    props.forEach((BoundElementOrDirectiveProperty prop) {
      if (!prop.isLiteral && isBlank(boundDirectivePropsIndex[prop.name])) {
        boundElementProps.add(this._createElementPropertyAst(
            elementName, prop.name, prop.expression, prop.sourceInfo));
      }
    });
    return boundElementProps;
  }

  BoundElementPropertyAst _createElementPropertyAst(
      String elementName, String name, AST ast, dynamic sourceInfo) {
    var unit = null;
    var bindingType;
    var boundPropertyName;
    var parts = StringWrapper.split(name, PROPERTY_PARTS_SEPARATOR);
    if (identical(parts.length, 1)) {
      boundPropertyName =
          this._schemaRegistry.getMappedPropName(dashCaseToCamelCase(parts[0]));
      bindingType = PropertyBindingType.Property;
      if (!this._schemaRegistry.hasProperty(elementName, boundPropertyName)) {
        this._reportError(
            '''Can\'t bind to \'${ boundPropertyName}\' since it isn\'t a known native property in ${ sourceInfo}''');
      }
    } else if (parts[0] == ATTRIBUTE_PREFIX) {
      boundPropertyName = dashCaseToCamelCase(parts[1]);
      bindingType = PropertyBindingType.Attribute;
    } else if (parts[0] == CLASS_PREFIX) {
      // keep original case!
      boundPropertyName = parts[1];
      bindingType = PropertyBindingType.Class;
    } else if (parts[0] == STYLE_PREFIX) {
      unit = parts.length > 2 ? parts[2] : null;
      boundPropertyName = dashCaseToCamelCase(parts[1]);
      bindingType = PropertyBindingType.Style;
    } else {
      this._reportError('''Invalid property name ${ name} in ${ sourceInfo}''');
      bindingType = null;
    }
    return new BoundElementPropertyAst(
        boundPropertyName, bindingType, ast, unit, sourceInfo);
  }

  List<String> _findComponentDirectiveNames(List<DirectiveAst> directives) {
    List<String> componentTypeNames = [];
    directives.forEach((directive) {
      var typeName = directive.directive.type.typeName;
      if (directive.directive.isComponent) {
        componentTypeNames.add(typeName);
      }
    });
    return componentTypeNames;
  }

  _assertOnlyOneComponent(List<DirectiveAst> directives, String sourceInfo) {
    var componentTypeNames = this._findComponentDirectiveNames(directives);
    if (componentTypeNames.length > 1) {
      this._reportError(
          '''More than one component: ${ componentTypeNames . join ( "," )} in ${ sourceInfo}''');
    }
  }

  _assertNoComponentsNorElementBindingsOnTemplate(
      List<DirectiveAst> directives,
      List<BoundElementPropertyAst> elementProps,
      List<BoundEventAst> events,
      String sourceInfo) {
    List<String> componentTypeNames =
        this._findComponentDirectiveNames(directives);
    if (componentTypeNames.length > 0) {
      this._reportError(
          '''Components on an embedded template: ${ componentTypeNames . join ( "," )} in ${ sourceInfo}''');
    }
    elementProps.forEach((prop) {
      this._reportError(
          '''Property binding ${ prop . name} not used by any directive on an embedded template in ${ prop . sourceInfo}''');
    });
    events.forEach((event) {
      this._reportError(
          '''Event binding ${ event . name} on an embedded template in ${ event . sourceInfo}''');
    });
  }
}

class BoundElementOrDirectiveProperty {
  String name;
  AST expression;
  bool isLiteral;
  String sourceInfo;
  BoundElementOrDirectiveProperty(
      this.name, this.expression, this.isLiteral, this.sourceInfo) {}
}

class ParseError {
  String message;
  String sourceInfo;
  ParseError(this.message, this.sourceInfo) {}
}

List<String> splitClasses(String classAttrValue) {
  return StringWrapper.split(classAttrValue.trim(), new RegExp(r'\s+'));
}

List<String> splitAtColon(String input, List<String> defaultValues) {
  var parts = StringWrapper.split(input.trim(), new RegExp(r'\s*:\s*'));
  if (parts.length > 1) {
    return parts;
  } else {
    return defaultValues;
  }
}
