library angular2.src.core.change_detection.change_detection;

import "jit_proto_change_detector.dart" show JitProtoChangeDetector;
import "pregen_proto_change_detector.dart" show PregenProtoChangeDetector;
import "proto_change_detector.dart" show DynamicProtoChangeDetector;
import "differs/iterable_differs.dart"
    show IterableDiffers, IterableDifferFactory;
import "differs/default_iterable_differ.dart" show DefaultIterableDifferFactory;
import "differs/keyvalue_differs.dart"
    show KeyValueDiffers, KeyValueDifferFactory;
import "differs/default_keyvalue_differ.dart" show DefaultKeyValueDifferFactory;
import "interfaces.dart"
    show
        ChangeDetection,
        ProtoChangeDetector,
        ChangeDetectorDefinition,
        ChangeDetectorGenConfig;
import "package:angular2/di.dart"
    show Injector, Inject, Injectable, OpaqueToken, Optional, Binding;
import "package:angular2/src/core/facade/collection.dart"
    show Map, StringMapWrapper;
import "package:angular2/src/core/facade/lang.dart"
    show isPresent, BaseException, assertionsEnabled;
export "parser/ast.dart"
    show
        ASTWithSource,
        AST,
        AstTransformer,
        PropertyRead,
        LiteralArray,
        ImplicitReceiver;
export "parser/lexer.dart" show Lexer;
export "parser/parser.dart" show Parser;
export "parser/locals.dart" show Locals;
export "exceptions.dart"
    show
        DehydratedException,
        ExpressionChangedAfterItHasBeenCheckedException,
        ChangeDetectionError;
export "interfaces.dart"
    show
        ProtoChangeDetector,
        ChangeDetector,
        ChangeDispatcher,
        ChangeDetection,
        ChangeDetectorDefinition,
        DebugContext,
        ChangeDetectorGenConfig;
export "constants.dart" show ChangeDetectionStrategy;
export "proto_change_detector.dart" show DynamicProtoChangeDetector;
export "binding_record.dart" show BindingRecord, BindingTarget;
export "directive_record.dart" show DirectiveIndex, DirectiveRecord;
export "dynamic_change_detector.dart" show DynamicChangeDetector;
export "change_detector_ref.dart" show ChangeDetectorRef;
export "differs/iterable_differs.dart"
    show IterableDiffers, IterableDiffer, IterableDifferFactory;
export "differs/keyvalue_differs.dart"
    show KeyValueDiffers, KeyValueDiffer, KeyValueDifferFactory;
export "pipe_transform.dart" show PipeTransform, PipeOnDestroy;
export "change_detection_util.dart" show WrappedValue;

/**
 * Structural diffing for `Object`s and `Map`s.
 */
const List<KeyValueDifferFactory> keyValDiff = const [
  const DefaultKeyValueDifferFactory()
];
/**
 * Structural diffing for `Iterable` types such as `Array`s.
 */
const List<IterableDifferFactory> iterableDiff = const [
  const DefaultIterableDifferFactory()
];
const defaultIterableDiffers = const IterableDiffers(iterableDiff);
const defaultKeyValueDiffers = const KeyValueDiffers(keyValDiff);
/**
 * Map from [ChangeDetectorDefinition#id] to a factory method which takes a
 * [Pipes] and a [ChangeDetectorDefinition] and generates a
 * [ProtoChangeDetector] associated with the definition.
 */

// TODO(kegluneq): Use PregenProtoChangeDetectorFactory rather than Function once possible in

// dart2js. See https://github.com/dart-lang/sdk/issues/23630 for details.
Map<String, Function> preGeneratedProtoDetectors = {};

/**
 * Implements change detection using a map of pregenerated proto detectors.
 */
@Injectable()
class PreGeneratedChangeDetection extends ChangeDetection {
  ChangeDetection _dynamicChangeDetection;
  Map<String, Function> _protoChangeDetectorFactories;
  ChangeDetectorGenConfig _genConfig;
  PreGeneratedChangeDetection(
      [ChangeDetectorGenConfig config,
      Map<String, Function> protoChangeDetectorsForTest])
      : super() {
    /* super call moved to initializer */;
    this._dynamicChangeDetection = new DynamicChangeDetection();
    this._protoChangeDetectorFactories = isPresent(protoChangeDetectorsForTest)
        ? protoChangeDetectorsForTest
        : preGeneratedProtoDetectors;
    this._genConfig = isPresent(config)
        ? config
        : new ChangeDetectorGenConfig(
            assertionsEnabled(), assertionsEnabled(), false);
  }
  static bool isSupported() {
    return PregenProtoChangeDetector.isSupported();
  }

  ProtoChangeDetector getProtoChangeDetector(
      String id, ChangeDetectorDefinition definition) {
    if (StringMapWrapper.contains(this._protoChangeDetectorFactories, id)) {
      return StringMapWrapper.get(
          this._protoChangeDetectorFactories, id)(definition);
    }
    return this._dynamicChangeDetection.getProtoChangeDetector(id, definition);
  }

  ChangeDetectorGenConfig get genConfig {
    return this._genConfig;
  }

  bool get generateDetectors {
    return true;
  }
}

/**
 * Implements change detection that does not require `eval()`.
 *
 * This is slower than [JitChangeDetection].
 */
@Injectable()
class DynamicChangeDetection extends ChangeDetection {
  ChangeDetectorGenConfig _genConfig;
  DynamicChangeDetection([ChangeDetectorGenConfig config]) : super() {
    /* super call moved to initializer */;
    this._genConfig = isPresent(config)
        ? config
        : new ChangeDetectorGenConfig(
            assertionsEnabled(), assertionsEnabled(), false);
  }
  ProtoChangeDetector getProtoChangeDetector(
      String id, ChangeDetectorDefinition definition) {
    return new DynamicProtoChangeDetector(definition);
  }

  ChangeDetectorGenConfig get genConfig {
    return this._genConfig;
  }

  bool get generateDetectors {
    return true;
  }
}

/**
 * Implements faster change detection by generating source code.
 *
 * This requires `eval()`. For change detection that does not require `eval()`, see
 * [DynamicChangeDetection] and [PreGeneratedChangeDetection].
 */
@Injectable()
class JitChangeDetection extends ChangeDetection {
  ChangeDetectorGenConfig _genConfig;
  JitChangeDetection([ChangeDetectorGenConfig config]) : super() {
    /* super call moved to initializer */;
    this._genConfig = isPresent(config)
        ? config
        : new ChangeDetectorGenConfig(
            assertionsEnabled(), assertionsEnabled(), false);
  }
  static bool isSupported() {
    return JitProtoChangeDetector.isSupported();
  }

  ProtoChangeDetector getProtoChangeDetector(
      String id, ChangeDetectorDefinition definition) {
    return new JitProtoChangeDetector(definition);
  }

  ChangeDetectorGenConfig get genConfig {
    return this._genConfig;
  }

  bool get generateDetectors {
    return true;
  }
}
