library angular2.src.router.location;

import "location_strategy.dart" show LocationStrategy;
import "package:angular2/src/core/facade/lang.dart"
    show StringWrapper, isPresent;
import "package:angular2/src/core/facade/async.dart"
    show EventEmitter, ObservableWrapper;
import "package:angular2/src/core/facade/lang.dart" show BaseException, isBlank;
import "package:angular2/di.dart"
    show OpaqueToken, Injectable, Optional, Inject;

const OpaqueToken APP_BASE_HREF = const OpaqueToken("appBaseHref");

/**
 * This is the service that an application developer will directly interact with.
 *
 * Responsible for normalizing the URL against the application's base href.
 * A normalized URL is absolute from the URL host, includes the application's base href, and has no
 * trailing slash:
 * - `/my/app/user/123` is normalized
 * - `my/app/user/123` **is not** normalized
 * - `/my/app/user/123/` **is not** normalized
 */
@Injectable()
class Location {
  LocationStrategy platformStrategy;
  EventEmitter _subject = new EventEmitter();
  String _baseHref;
  Location(this.platformStrategy,
      [@Optional() @Inject(APP_BASE_HREF) String href]) {
    var browserBaseHref =
        isPresent(href) ? href : this.platformStrategy.getBaseHref();
    if (isBlank(browserBaseHref)) {
      throw new BaseException(
          '''No base href set. Either provide a binding to "appBaseHrefToken" or add a base element.''');
    }
    this._baseHref = stripTrailingSlash(stripIndexHtml(browserBaseHref));
    this.platformStrategy.onPopState((_) {
      ObservableWrapper.callNext(
          this._subject, {"url": this.path(), "pop": true});
    });
  }
  String path() {
    return this.normalize(this.platformStrategy.path());
  }

  String normalize(String url) {
    return stripTrailingSlash(
        _stripBaseHref(this._baseHref, stripIndexHtml(url)));
  }

  String normalizeAbsolutely(String url) {
    if (!url.startsWith("/")) {
      url = "/" + url;
    }
    return stripTrailingSlash(_addBaseHref(this._baseHref, url));
  }

  void go(String url) {
    var finalUrl = this.normalizeAbsolutely(url);
    this.platformStrategy.pushState(null, "", finalUrl);
  }

  void forward() {
    this.platformStrategy.forward();
  }

  void back() {
    this.platformStrategy.back();
  }

  void subscribe(dynamic /* (value: any) => void */ onNext,
      [dynamic /* (exception: any) => void */ onThrow = null,
      dynamic /* () => void */ onReturn = null]) {
    ObservableWrapper.subscribe(this._subject, onNext, onThrow, onReturn);
  }
}

String _stripBaseHref(String baseHref, String url) {
  if (baseHref.length > 0 && url.startsWith(baseHref)) {
    return url.substring(baseHref.length);
  }
  return url;
}

String _addBaseHref(String baseHref, String url) {
  if (!url.startsWith(baseHref)) {
    return baseHref + url;
  }
  return url;
}

String stripIndexHtml(String url) {
  if (new RegExp(r'\/index.html$').hasMatch(url)) {
    // '/index.html'.length == 11
    return url.substring(0, url.length - 11);
  }
  return url;
}

String stripTrailingSlash(String url) {
  if (new RegExp(r'\/$').hasMatch(url)) {
    url = url.substring(0, url.length - 1);
  }
  return url;
}
